@@ -36,6 +36,7 @@ import pyghmi.ipmi.private.util as pygutil
 import pyghmi.ipmi.sdr as sdr
 import socket
 import struct
+import threading
 
 try:
     range = xrange
@@ -95,6 +96,19 @@ def _mask_to_cidr(mask):
 def _cidr_to_mask(prefix):
     return struct.pack('>I', 2 ** prefix - 1 << (32 - prefix))
 
+class Housekeeper(threading.Thread):
+    """A Maintenance thread for housekeeping
+
+    Long lived use of pyghmi may warrant some recurring asynchronous behavior.
+    This stock thread provides a simple minimal context for these housekeeping
+    tasks to run in.  To use, do 'pyghmi.ipmi.command.Maintenance().start()'
+    and from that point forward, pyghmi should execute any needed ongoing
+    tasks automatically as needed.  This is an alternative to calling
+    wait_for_rsp or eventloop in a thread of the callers design.
+    """
+    def run(self):
+        Command.eventloop()
+
 
 class Command(object):
     """Send IPMI commands to BMCs.
@@ -116,19 +130,26 @@ class Command(object):
     :param onlogon: function to run when logon completes in an asynchronous
                     fashion.  This will result in a greenthread behavior.
     :param kg: Optional parameter to use if BMC has a particular Kg configured
+    :param verifycallback:  For OEM extensions that use HTTPS, this function
+                            will be used to evaluate the certificate.
+    :param keepalive:  If False, then an idle connection will logout rather than keepalive
+                       unless held open by console or ongoing activity.
     """
 
     def __init__(self, bmc=None, userid=None, password=None, port=623,
-                 onlogon=None, kg=None):
+                 onlogon=None, kg=None, privlevel=None, verifycallback=None,
+                 keepalive=True):
         # TODO(jbjohnso): accept tuples and lists of each parameter for mass
         # operations without pushing the async complexities up the stack
         self.onlogon = onlogon
         self.bmc = bmc
+        self._sdrcachedir = None
         self._sdr = None
         self._oem = None
+        self._oemknown = False
         self._netchannel = None
         self._ipv6support = None
-        self.certverify = None
+        self.certverify = verifycallback
         if bmc is None:
             self.ipmi_session = localsession.Session()
         elif onlogon is not None:
@@ -137,7 +158,9 @@ class Command(object):
                                                 password=password,
                                                 onlogon=self.logged,
                                                 port=port,
-                                                kg=kg)
+                                                kg=kg,
+                                                privlevel=privlevel,
+                                                keepalive=keepalive)
             # induce one iteration of the loop, now that we would be
             # prepared for it in theory
             session.Session.wait_for_rsp(0)
@@ -146,7 +169,19 @@ class Command(object):
                                                 userid=userid,
                                                 password=password,
                                                 port=port,
-                                                kg=kg)
+                                                kg=kg,
+                                                privlevel=privlevel)
+
+    def set_sdr_cachedir(self, path):
+        """Register use of a directory for SDR cache.
+
+        Takes the given directory and uses it to persist SDR cache run to run.
+        This can greatly improve performance across runs.
+
+        :param path:
+        :return:
+        """
+        self._sdrcachedir = path
 
     def register_key_handler(self, callback, type='tls'):
         """Assign a verification handler for a public key
@@ -170,8 +205,8 @@ class Command(object):
 
     @classmethod
     def eventloop(cls):
-        while session.Session.wait_for_rsp():
-            pass
+        while True:
+            session.Session.wait_for_rsp()
 
     @classmethod
     def wait_for_rsp(cls, timeout):
@@ -204,9 +239,10 @@ class Command(object):
         method does an interrogation to identify the OEM.
 
         """
-        if self._oem:
+        if self._oemknown:
             return
-        self._oem = get_oem_handler(self._get_device_id(), self)
+        self._oem, self._oemknown = get_oem_handler(self._get_device_id(),
+                                                    self)
 
     def get_bootdev(self):
         """Get current boot device override information.
@@ -412,9 +448,23 @@ class Command(object):
                                             retry=retry, timeout=timeout)
         if 'error' in rsp:
             raise exc.IpmiException(rsp['error'], rsp['code'])
-        rsp['data'] = buffer(bytearray(rsp['data']))
+        rsp['data'] = buffer(rsp['data'])
         return rsp
 
+    def get_diagnostic_data(self, savefile, progress=None):
+        self.oem_init()
+        return self._oem.get_diagnostic_data(savefile, progress)
+
+    def get_description(self):
+        """Get physical attributes for the system, e.g. for GUI use
+
+        :returns: dict -- dict containing attributes, 'height' is for
+               how many U tall, 'slot' for what slot in a blade enclosure
+               or 0 if not blade, for example.
+        """
+        self.oem_init()
+        return self._oem.get_description()
+
     def raw_command(self, netfn, command, bridge_request=(), data=(),
                     delay_xmit=None, retry=True, timeout=None):
         """Send raw ipmi command to BMC
@@ -434,10 +484,13 @@ class Command(object):
         :param timeout: A custom amount of time to wait for initial reply
         :returns: dict -- The response from IPMI device
         """
-        return self.ipmi_session.raw_command(netfn=netfn, command=command,
-                                             bridge_request=bridge_request,
-                                             data=data, delay_xmit=delay_xmit,
-                                             retry=retry, timeout=timeout)
+        rsp = self.ipmi_session.raw_command(netfn=netfn, command=command,
+                                            bridge_request=bridge_request,
+                                            data=data, delay_xmit=delay_xmit,
+                                            retry=retry, timeout=timeout)
+        if 'data' in rsp:
+            rsp['data'] = list(rsp['data'])
+        return rsp
 
     def get_power(self):
         """Get current power state of the managed system
@@ -506,7 +559,7 @@ class Command(object):
         # sdr timestamp, our data version revision, aux firmware revision,
         # and oem defined field
         if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+            self._sdr = sdr.SDR(self, self._sdrcachedir)
         return self._sdr
 
     def get_event_log(self, clear=False):
@@ -553,8 +606,7 @@ class Command(object):
         returns an iterable of string descriptions
         """
         yield "System"
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for fruid in sorted(self._sdr.fru):
             yield self._sdr.fru[fruid].fru_name
         self.oem_init()
@@ -570,12 +622,12 @@ class Command(object):
         self.oem_init()
         if component == 'System':
             return self._get_zero_fru()
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for fruid in self._sdr.fru:
             if self._sdr.fru[fruid].fru_name == component:
                 return self._oem.process_fru(fru.FRU(
-                    ipmicmd=self, fruid=fruid, sdr=self._sdr.fru[fruid]).info)
+                    ipmicmd=self, fruid=fruid, sdr=self._sdr.fru[fruid]).info,
+                                             component)
         return self._oem.get_inventory_of_component(component)
 
     def _get_zero_fru(self):
@@ -614,13 +666,13 @@ class Command(object):
         """
         self.oem_init()
         yield ("System", self._get_zero_fru())
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for fruid in sorted(self._sdr.fru):
             fruinf = fru.FRU(
                 ipmicmd=self, fruid=fruid, sdr=self._sdr.fru[fruid]).info
             if fruinf is not None:
-                fruinf = self._oem.process_fru(fruinf)
+                fruinf = self._oem.process_fru(fruinf,
+                                               self._sdr.fru[fruid].fru_name)
             yield (self._sdr.fru[fruid].fru_name, fruinf)
         for componentpair in self._oem.get_oem_inventory():
             yield componentpair
@@ -657,10 +709,18 @@ class Command(object):
         warning, critical, or failed assessments.
         """
         summary = {'badreadings': [], 'health': const.Health.Ok}
-        for reading in self.get_sensor_data():
-            if reading.health != const.Health.Ok:
-                summary['health'] |= reading.health
-                summary['badreadings'].append(reading)
+        fallbackreadings = []
+        try:
+            self.oem_init()
+            fallbackreadings = self._oem.get_health(summary)
+            for reading in self.get_sensor_data():
+                if reading.health != const.Health.Ok:
+                    summary['health'] |= reading.health
+                    summary['badreadings'].append(reading)
+        except exc.BypassGenericBehavior:
+            pass
+        if not summary['badreadings']:
+            summary['badreadings'] = fallbackreadings
         return summary
 
     def get_sensor_reading(self, sensorname):
@@ -672,8 +732,7 @@ class Command(object):
         :param sensorname:  Name of the desired sensor
         :returns: sdr.SensorReading object
         """
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for sensor in self._sdr.get_sensor_numbers():
             if self._sdr.sensors[sensor].name == sensorname:
                 rsp = self.raw_command(command=0x2d, netfn=4, data=(sensor,))
@@ -715,14 +774,22 @@ class Command(object):
         else:
             raise Exception("Unrecognized data format " + repr(fetchdata))
 
-    def get_system_configuration(self):
+    def get_system_configuration(self, hideadvanced=True):
         self.oem_init()
-        return self._oem.get_system_configuration()
+        return self._oem.get_system_configuration(hideadvanced)
 
     def set_system_configuration(self, changeset):
         self.oem_init()
         return self._oem.set_system_configuration(changeset)
 
+    def clear_system_configuration(self):
+        """Clear the Bios/UEFI configuration
+
+        This requests the system revert to factory default settings
+        """
+        self.oem_init()
+        self._oem.clear_system_configuration()
+
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, channel=None):
         """Set network configuration data.
@@ -862,8 +929,7 @@ class Command(object):
 
         :returns: Iterator of sdr.SensorReading objects
         """
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for sensor in self._sdr.get_sensor_numbers():
             rsp = self.raw_command(command=0x2d, netfn=4, data=(sensor,))
             if 'error' in rsp:
@@ -882,8 +948,7 @@ class Command(object):
 
         :returns: Iterator of dicts describing each sensor
         """
-        if self._sdr is None:
-            self._sdr = sdr.SDR(self)
+        self.init_sdr()
         for sensor in self._sdr.get_sensor_numbers():
             yield {'name': self._sdr.sensors[sensor].name,
                    'type': self._sdr.sensors[sensor].sensor_type}
@@ -1137,6 +1202,23 @@ class Command(object):
         if not ip == '0.0.0.0':
             self._assure_alert_policy(channel, destination)
 
+    def get_hostname(self):
+        """Get the hostname used by the BMC in various contexts
+
+        This can vary somewhat in interpretation, but generally speaking
+        this should be the name that shows up on UIs and in DHCP requests and
+        DNS registration requests, as applicable.
+
+        :return: current hostname
+        """
+        self.oem_init()
+        try:
+            return self._oem.get_hostname()
+        except exc.UnsupportedFunctionality:
+            # Use the DCMI MCI field as a fallback, since it's the closest
+            # thing in the IPMI Spec for this
+            return self.get_mci()
+
     def get_mci(self):
         """Get the Management Controller Identifier, per DCMI specification
 
@@ -1144,6 +1226,20 @@ class Command(object):
         """
         return self._chunkwise_dcmi_fetch(9)
 
+    def set_hostname(self, hostname):
+        """Set the hostname to be used by the BMC in various contexts.
+
+        See get_hostname for details
+
+        :param hostname: The hostname to set
+        :return: Nothing
+        """
+        self.oem_init()
+        try:
+            return self._oem.set_hostname(hostname)
+        except exc.UnsupportedFunctionality:
+            return self.set_mci(hostname)
+
     def set_mci(self, mci):
         """Set the management controller identifier, per DCMI specification
 
@@ -1499,6 +1595,9 @@ class Command(object):
             'proprietary': 5,
             'no_access': 0x0F,
         }
+        self.oem_init()
+        self._oem.set_user_access(
+            uid, channel, callback, link_auth, ipmi_msg, privilege_level)
         data = [b, uid & 0b00111111,
                 privilege_levels[privilege_level] & 0b00001111, 0]
         response = self.raw_command(netfn=0x06, command=0x43, data=data)
@@ -1666,13 +1765,17 @@ class Command(object):
                 ipmi_msg (bool)
                 privilege_level: (str)[callback, user, operatorm administrator,
                                        proprietary, no_access]
+            expiration:
+                None for 'unknown', 0 for no expiry, days to expire otherwise.
         """
         if channel is None:
             channel = self.get_network_channel()
         name = self.get_user_name(uid)
         access = self.get_user_access(uid, channel)
+        self.oem_init()
+        expiration = self._oem.get_user_expiration(uid)
         data = {'name': name, 'uid': uid, 'channel': channel,
-                'access': access['access']}
+                'access': access['access'], 'expiration': expiration}
         return data
 
     def get_name_uids(self, name, channel=None):
@@ -1737,11 +1840,11 @@ class Command(object):
         if channel is None:
             channel = self.get_network_channel()
         self.set_user_name(uid, name)
+        self.set_user_password(uid, password=password)
+        self.set_user_password(uid, mode='enable', password=password)
         self.set_user_access(uid, channel, callback=callback,
                              link_auth=link_auth, ipmi_msg=ipmi_msg,
                              privilege_level=privilege_level)
-        self.set_user_password(uid, password=password)
-        self.set_user_password(uid, mode='enable', password=password)
         return True
 
     def user_delete(self, uid, channel=None):
@@ -1883,7 +1986,7 @@ class Command(object):
 
         :param filename: The filename to use, the basename of the parameter
                          will be given to the bmc.
-        :param filename: Optional callback for progress updates
+        :param progress: Optional callback for progress updates
         """
         self.oem_init()
         return self._oem.upload_media(filename, progress)
@@ -1895,3 +1998,19 @@ class Command(object):
         """
         self.oem_init()
         return self._oem.list_media()
+
+    def get_licenses(self):
+        self.oem_init()
+        return self._oem.get_licenses()
+
+    def delete_license(self, name):
+        self.oem_init()
+        return self._oem.delete_license(name)
+
+    def save_licenses(self, directory):
+        self.oem_init()
+        return self._oem.save_licenses(directory)
+
+    def apply_license(self, filename, progress=None):
+        self.oem_init()
+        return self._oem.apply_license(filename, progress)
