@@ -1,6 +1,7 @@
 # vim: tabstop=4 shiftwidth=4 softtabstop=4
 
 # Copyright 2014 IBM Corporation
+# Copyright 2015-2019 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -22,10 +23,7 @@ import threading
 
 from pyghmi.ipmi.private import constants
 from pyghmi.ipmi.private import session
-from pyghmi.ipmi.private import util
-
-
-PENDINGOUTPUT = threading.RLock()
+from pyghmi.ipmi.private.util import _monotonic_time
 
 
 class Console(object):
@@ -47,6 +45,7 @@ class Console(object):
     def __init__(self, bmc, userid, password,
                  iohandler, port=623,
                  force=False, kg=None):
+        self.outputlock = threading.RLock()
         self.keepaliveid = None
         self.connected = False
         self.broken = False
@@ -59,6 +58,9 @@ class Console(object):
         self.awaitingack = False
         self.activated = False
         self.force_session = force
+        self.port = port
+        self.ipmi_session = None
+        self.callgotsession = None
         self.ipmi_session = session.Session(bmc=bmc,
                                             userid=userid,
                                             password=password,
@@ -68,6 +70,9 @@ class Console(object):
         # induce one iteration of the loop, now that we would be
         # prepared for it in theory
         session.Session.wait_for_rsp(0)
+        if self.callgotsession is not None:
+            self._got_session(self.callgotsession)
+            self.callgotsession = None
 
     def _got_session(self, response):
         """Private function to navigate SOL payload activation
@@ -75,6 +80,9 @@ class Console(object):
         if 'error' in response:
             self._print_error(response['error'])
             return
+        if not self.ipmi_session:
+            self.callgotsession = response
+            return
         # Send activate sol payload directive
         # netfn= 6 (application)
         # command = 0x48 (activate payload)
@@ -129,7 +137,10 @@ class Console(object):
         # BMC tells us this is the maximum allowed size
         # data[6:7] is the promise of how small packets are going to be, but we
         # don't have any reason to worry about it
-        if (data[8] + (data[9] << 8)) not in (623, 28418):
+        # some BMCs disagree on the endianness, so do both
+        valid_ports = (self.port, struct.unpack(
+            '<H', struct.pack('>H', self.port))[0])
+        if (data[8] + (data[9] << 8)) not in valid_ports:
             # TODO(jbjohnso): support atypical SOL port number
             raise NotImplementedError("Non-standard SOL Port Number")
         # ignore data[10:11] for now, the vlan detail, shouldn't matter to this
@@ -167,16 +178,16 @@ class Console(object):
         # discern which channel or even *if* the serial port in question
         # correlates at all to an ipmi channel to check mux
 
-    @util.protect(PENDINGOUTPUT)
     def _addpendingdata(self, data):
-        if isinstance(data, dict):
-            self.pendingoutput.append(data)
-        else:  # it is a text situation
-            if (len(self.pendingoutput) == 0 or
-                    isinstance(self.pendingoutput[-1], dict)):
+        with self.outputlock:
+            if isinstance(data, dict):
                 self.pendingoutput.append(data)
-            else:
-                self.pendingoutput[-1] += data
+            else:  # it is a text situation
+                if (len(self.pendingoutput) == 0 or
+                        isinstance(self.pendingoutput[-1], dict)):
+                    self.pendingoutput.append(data)
+                else:
+                    self.pendingoutput[-1] += data
 
     def _got_cons_input(self, handle):
         """Callback for handle events detected by ipmi session
@@ -225,24 +236,27 @@ class Console(object):
         """
         return session.Session.wait_for_rsp(timeout=timeout)
 
-    @util.protect(PENDINGOUTPUT)
     def _sendpendingoutput(self):
-        if len(self.pendingoutput) == 0:
-            return
-        if isinstance(self.pendingoutput[0], dict):
-            if 'break' in self.pendingoutput[0]:
-                self._sendoutput("", sendbreak=True)
+        with self.outputlock:
+            dobreak = False
+            chunk = ''
+            if len(self.pendingoutput) == 0:
+                return
+            if isinstance(self.pendingoutput[0], dict):
+                if 'break' in self.pendingoutput[0]:
+                    dobreak = True
+                else:
+                    del self.pendingoutput[0]
+                    raise ValueError
+                del self.pendingoutput[0]
+            elif len(self.pendingoutput[0]) > self.maxoutcount:
+                chunk = self.pendingoutput[0][:self.maxoutcount]
+                self.pendingoutput[0] = self.pendingoutput[0][
+                    self.maxoutcount:]
             else:
-                raise ValueError
-            del self.pendingoutput[0]
-            return
-        if len(self.pendingoutput[0]) > self.maxoutcount:
-            chunk = self.pendingoutput[0][:self.maxoutcount]
-            self.pendingoutput[0] = self.pendingoutput[0][self.maxoutcount:]
-        else:
-            chunk = self.pendingoutput[0]
-            del self.pendingoutput[0]
-        self._sendoutput(chunk)
+                chunk = self.pendingoutput[0]
+                del self.pendingoutput[0]
+            self._sendoutput(chunk, sendbreak=dobreak)
 
     def _sendoutput(self, output, sendbreak=False):
         self.myseq += 1
@@ -255,22 +269,36 @@ class Console(object):
         breakbyte = 0
         if sendbreak:
             breakbyte = 0b10000
-        payload = struct.pack("BBBB", self.myseq, 0, 0, breakbyte)
-        payload += output
+        try:
+            payload = bytearray((self.myseq, 0, 0, breakbyte)) + output
+        except TypeError:  # bytearray hits unicode...
+            payload = bytearray((self.myseq, 0, 0, breakbyte
+                                 )) + output.encode('utf8')
         self.lasttextsize = len(output)
         needskeepalive = False
         if self.lasttextsize == 0:
             needskeepalive = True
         self.awaitingack = True
-        payload = struct.unpack("%dB" % len(payload), payload)
         self.lastpayload = payload
-        self.send_payload(payload, needskeepalive=needskeepalive)
+        self.send_payload(payload, retry=False, needskeepalive=needskeepalive)
+        retries = 5
+        while retries and self.awaitingack:
+            expiry = _monotonic_time() + 5.5 - retries
+            while self.awaitingack and _monotonic_time() < expiry:
+                self.wait_for_rsp(0.5)
+            if self.awaitingack:
+                self.send_payload(payload, retry=False,
+                                  needskeepalive=needskeepalive)
+            retries -= 1
+        if not retries:
+            self._print_error('Connection lost')
 
     def send_payload(self, payload, payload_type=1, retry=True,
                      needskeepalive=False):
         while not (self.connected or self.broken):
             session.Session.wait_for_rsp(timeout=10)
         if not self.ipmi_session.logged:
+            self._print_error('Session no longer connected')
             raise exc.IpmiException('Session no longer connected')
         self.ipmi_session.send_payload(payload,
                                        payload_type=payload_type,
@@ -329,10 +357,10 @@ class Console(object):
                 remdatalen = len(payload[4:])  # store remote len before dupe
                 # retry logic, we must ack *this* many even if it is
                 # a retry packet with new partial data
-                remdata = struct.pack("%dB" % remdatalen, *payload[4:])
+                remdata = bytes(payload[4:])
             if newseq == self.remseq:  # it is a retry, but could have new data
                 if remdatalen > self.lastsize:
-                    remdata = remdata[4 + self.lastsize:]
+                    remdata = bytes(remdata[4 + self.lastsize:])
                 else:  # no new data...
                     remdata = ""
             else:  # TODO(jbjohnso) what if remote sequence number is wrong??
@@ -340,7 +368,7 @@ class Console(object):
             self.lastsize = remdatalen
             if remdata:  # Do not subject callers to empty data
                 self._print_data(remdata)
-            ackpayload = (0, self.remseq, remdatalen, 0)
+            ackpayload = bytearray((0, self.remseq, remdatalen, 0))
             # Why not put pending data into the ack? because it's rare
             # and might be hard to decide what to do in the context of
             # retry situation
@@ -361,8 +389,7 @@ class Console(object):
                 else:  # retry all or part of packet, but in a new form
                     # also add pending output for efficiency and ease
                     newtext = self.lastpayload[4 + ackcount:]
-                    newtext = struct.pack("B"*len(newtext), *newtext)
-                    with util.protect(PENDINGOUTPUT):
+                    with self.outputlock:
                         if (self.pendingoutput and
                                 not isinstance(self.pendingoutput[0], dict)):
                             self.pendingoutput[0] = \
@@ -380,7 +407,7 @@ class Console(object):
             # try to mitigate by avoiding overeager retries
             # occasional retry of a packet
             # sooner than timeout suggests is evidently a big deal
-            self.send_payload(payload=self.lastpayload)
+            self.send_payload(payload=self.lastpayload, retry=False)
 
     def main_loop(self):
         """Process all events until no more sessions exist.
@@ -409,6 +436,7 @@ class ServerConsole(Console):
     """
 
     def __init__(self, _session, iohandler, force=False):
+        self.outputlock = threading.RLock()
         self.keepaliveid = None
         self.connected = True
         self.broken = False
@@ -458,16 +486,16 @@ class ServerConsole(Console):
                 remdatalen = len(payload[4:])  # store remote len before dupe
                 # retry logic, we must ack *this* many even if it is
                 # a retry packet with new partial data
-                remdata = struct.pack("%dB" % remdatalen, *payload[4:])
+                remdata = bytes(payload[4:])
             if newseq == self.remseq:  # it is a retry, but could have new data
                 if remdatalen > self.lastsize:
-                    remdata = remdata[4 + self.lastsize:]
+                    remdata = bytes(remdata[4 + self.lastsize:])
                 else:  # no new data...
                     remdata = ""
             else:  # TODO(jbjohnso) what if remote sequence number is wrong??
                 self.remseq = newseq
             self.lastsize = remdatalen
-            ackpayload = (0, self.remseq, remdatalen, flag)
+            ackpayload = bytearray((0, self.remseq, remdatalen, flag))
             # Why not put pending data into the ack? because it's rare
             # and might be hard to decide what to do in the context of
             # retry situation
@@ -483,8 +511,7 @@ class ServerConsole(Console):
             self.awaitingack = False
             if nacked and not breakdetected:  # the BMC was in some way unhappy
                 newtext = self.lastpayload[4 + ackcount:]
-                newtext = struct.pack("B"*len(newtext), *newtext)
-                with util.protect(PENDINGOUTPUT):
+                with self.outputlock:
                     if (self.pendingoutput and
                             not isinstance(self.pendingoutput[0], dict)):
                         self.pendingoutput[0] = newtext + self.pendingoutput[0]
